/*
	SHARP PC-G800 series emulator
	C
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#define DEF_GLOBAL
#include "g800.h"

/* AhX0000~003f̏l */
static uint8 ram_0000_003f[] = {
	0xc9, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xc9, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xc9, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xc9, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xc9, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xc9, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xc9, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xc9, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

/*
	ROM1y[Wǂݍ ()
*/
static size_t loadROM1page(uint8 *p, const char *dir, int page, size_t size)
{
	size_t result;
	char path[FILENAME_MAX];
	
	if(page < 0)
		sprintf(path, "%s/base.bin", dir);
	else
		sprintf(path, "%s/rom%02x.bin", dir, page);
	if((result = readBin(path, p, size)) > 0)
		return result;

	if(page < 0)
		sprintf(path, "%s/base.txt", dir);
	else
		sprintf(path, "%s/rom%02x.txt", dir, page);
	if((result = readHexAbs(path, p, NULL, size, FALSE)) > 0)
		return result;
	return 0;
}

/*
	ROMǂݍ ()
*/
static int loadROM(const char *dir)
{
	int page;

	if(!useROM)
		return TRUE;

	/* 0000-003fǂݍ */
	if(loadROM1page(memory, dir, -1, 0x40) < 0x40)
		return FALSE;
	memcpy(ram_0000_003f, memory, sizeof(ram_0000_003f));

	/* ROM̃y[W𒲂ׂ */
	for(romBanks = 0; loadROM1page(NULL, dir, romBanks, 0x4000) == 0x4000; romBanks++)
		;
	if(romBanks == 0)
		return FALSE;

	/* ROMǂݍ */
	rom = malloc(0x4000 * romBanks);
	for(page = 0; page != romBanks; page++)
		loadROM1page(ROM(page), dir, page, 0x4000);

	/* BANK */
	romBank = exBank = 0;
	memcpy(&memory[0x8000], ROM(0), 0x4000);
	memcpy(&memory[0xc000], ROM(0), 0x4000);
	return TRUE;
}

/*
	RAMǂݍ ()
*/
static int loadRAM(const char *path)
{
	return (readHex(path, memory, NULL, 0x8000, FALSE) > 0);
}

/*
	RAMۑ ()
*/
static int storeRAM(const char *path)
{
	return (writeHex(path, memory, 0x40, 0x8000 - 0x40) > 0);
}

/*
	Zbg ()
*/
static void reset(void)
{
	z80srand(time(NULL));
	z80reset(&z80);
	memcpy(memory, ram_0000_003f, sizeof(ram_0000_003f));
	memset(keyMatrix, 0, sizeof(keyMatrix));
	keyBreak = keyShift = keyReset = 0;
}

/*
	G~[g ()
*/
static void boot(void)
{
	z80outport(NULL, 0x11, 0);
	z80outport(NULL, 0x12, 0);
	z80outport(NULL, 0x14, 0);
	z80outport(NULL, 0x15, 1);
	z80outport(NULL, 0x16, 0xff);
	z80outport(NULL, 0x17, 0xf);
	z80outport(NULL, 0x18, 0);
	z80outport(NULL, 0x19, 0);
	z80outport(NULL, 0x1b, 0);
	z80outport(NULL, 0x1c, 1);
	z80.r.im = 1;
	memory[0x790d] = 0;

	switch(machine) {
	case MACHINE_E200:
		z80outport(NULL, 0x58, 0xc0);
		break;
	case MACHINE_G815:
		z80outport(NULL, 0x50, 0xc0);
		break;
	case MACHINE_G850:
		memory[0x779c] = (memory[0x779c] >= 0x07 && memory[0x779c] <= 0x1f ? memory[0x779c]: 0x0f);
		z80outport(NULL, 0x40, 0x24);
		z80outport(NULL, 0x40, memory[0x790d] + 0x40);
		z80outport(NULL, 0x40, memory[0x779c] + 0x80);
		z80outport(NULL, 0x40, 0xa0);
		z80outport(NULL, 0x40, 0xa4);
		z80outport(NULL, 0x40, 0xa6);
		z80outport(NULL, 0x40, 0xa9);
		z80outport(NULL, 0x40, 0xaf);
		z80outport(NULL, 0x40, 0xc0);
		z80outport(NULL, 0x40, 0x25);
		z80outport(NULL, 0x60, 0);
		z80outport(NULL, 0x61, 0xff);
		z80outport(NULL, 0x62, 0);
		z80outport(NULL, 0x64, 0);
		z80outport(NULL, 0x65, 1);
		z80outport(NULL, 0x66, 1);
		z80outport(NULL, 0x67, 0);
		z80outport(NULL, 0x6b, 4);
		z80outport(NULL, 0x6c, 0);
		z80outport(NULL, 0x6d, 0);
		z80outport(NULL, 0x6e, 4);
		break;
	}
}

/*
	vOs
*/
int exec(int exit_when_underflow)
{
	int wait, count_timer, states_update_io, init_timer;
	uint8 itype;

	wait = 1000 / freqUpdateIO;
	count_timer = init_timer = freqUpdateIO * 45 / 100;
	z80.i.stack_under = (exit_when_underflow ? 0x7ff6: 0xffff);
	z80.i.states = 0;

	for(;;) {
		/* EFCg */
		states_update_io = freqCPU / (csClk ? 2: 1) / freqUpdateIO;
		do {
			z80.i.states        += states_update_io;
			statesKeyStrobeLast += states_update_io;
			delay(wait);
		} while(z80.i.states < 0);

		/* TEhobt@؂ւ */
		if(buzzer != BUZZER_NONE)
			flipSoundBuffer();

		/* R[hs */
		switch(z80exec(&z80)) {
		case Z80_RUN:
			break;
		case Z80_HALT:
			if(ioReset != 0 || z80.r.iff != 0)
				break;
		case Z80_UNDERFLOW:
			updateLCD();
			return 0;
		}

		/* L[XVEL[荞 */
		itype = updateKey();
		if(itype & interruptMask) {
			interruptType |= itype;
			z80int1(&z80);
		}

		/* ^C}XVE^C}荞 */
		if(count_timer-- == 0) {
			count_timer = init_timer;
			if(interruptMask & INTERRUPT_1S) {
				timer ^= TIMER_1S;
				interruptType |= INTERRUPT_1S;
				z80int1(&z80);
			}
		}

		/* ZbgL[ */
		if(keyReset) {
			reset();
			boot();
			if(!useROM)
				return 0;
		}

		/* LCDXV */
		updateLCD();
	}
}

int main(int argc, char *argv[])
{
	/*  */
	if(!init(argc, argv))
		popup("!", "CANNOT OPEN CONFIG FILE");
	if(!loadROM(dirROM))
		popup("!", "CANNOT OPEN ROM FILES");
	loadRAM(pathRAM);
	if(pathProg != NULL && strcmp(pathProg, "") != 0)
		if(loadProg(NULL, pathProg) < 0) {
			popup("!", "CANNOT OPEN %s", pathProg);
			beginProg = 0;
		}
	reset();
	boot();

	/* s */
	if(beginProg != 0) {
		/* AhXw肵Ďs */
		go(beginProg);
		exec(TRUE);
	} else if(useROM) {
		/* 擪s */
		exec(FALSE);
	} else {
		/* [j^N */
		monitor();
	}

	/* ㏈ */
	storeRAM(pathRAM);
	return 0;
}

/*
	Copyright 2005 ~ 2008 maruhiro
	All rights reserved. 

	Redistribution and use in source and binary forms, 
	with or without modification, are permitted provided that 
	the following conditions are met: 

	 1. Redistributions of source code must retain the above copyright notice, 
	    this list of conditions and the following disclaimer. 

	 2. Redistributions in binary form must reproduce the above copyright notice, 
	    this list of conditions and the following disclaimer in the documentation 
	    and/or other materials provided with the distribution. 

	THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
	FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
	THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* eof */
