/*
	SHARP PC-G800 series Emulator
	IOCSG~[g
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include "g800.h"
#include "font.h"

/* XN[rowVRAMrowɕϊ */
#define LCD_TOP()	(machine == MACHINE_G850 ? memory[0x790d]: 0)
#define ROW_LCD2VRAM(row)	(((row) + LCD_TOP()) % vramRows)

/* L[R[h -> ASCIIR[hϊe[u */
const static uint8 keycode2ascii_l[] = {
	0x00, 0x06, 0x51, 0x57, 0x45, 0x52, 0x54, 0x59,
	0x55, 0x41, 0x53, 0x44, 0x46, 0x47, 0x48, 0x4a,
	0x4b, 0x5a, 0x58, 0x43, 0x56, 0x42, 0x4e, 0x4d,
	0x2c, 0x01, 0x02, 0x14, 0x11, 0x0a, 0x20, 0x1f,
	0x1e, 0x1d, 0x1c, 0x15, 0x30, 0x2e, 0x3d, 0x2b,
	0x0d, 0x4c, 0x3b, 0x17, 0x31, 0x32, 0x33, 0x2d,
	0x1a, 0x49, 0x4f, 0x12, 0x34, 0x35, 0x36, 0x2a,
	0x19, 0x50, 0x08, 0xfe, 0x37, 0x38, 0x39, 0x2f,
	0x29, 0xfe, 0xfe, 0xfe, 0xfe, 0x3e, 0x28, 0xfe,
	0xfe, 0x10, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0x0f,
	0x0c, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x06, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26,
	0x27, 0x5b, 0x5d, 0x7b, 0x7d, 0x5c, 0x7c, 0x7e,
	0x5f, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe,
	0x3f, 0xf0, 0x03, 0x14, 0x11, 0x0a, 0x20, 0x1f,
	0x1e, 0x1d, 0x1c, 0xf2, 0x30, 0x13, 0x45, 0x2b,
	0x07, 0x3d, 0x3a, 0x18, 0x31, 0x32, 0x33, 0x16,
	0x1b, 0x3c, 0x3e, 0x09, 0x34, 0x35, 0x36, 0x2a,
	0x19, 0x40, 0x08, 0xfe, 0xdf, 0x27, 0xf8, 0x2f,
	0xf1, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe,
	0x04, 0x10, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0x0e,
	0x0b, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};
const static uint8 keycode2ascii_s[] = {
	0x00, 0x06, 0x71, 0x77, 0x65, 0x72, 0x74, 0x79,
	0x75, 0x61, 0x73, 0x64, 0x66, 0x67, 0x68, 0x6a,
	0x6b, 0x7a, 0x78, 0x63, 0x76, 0x62, 0x6e, 0x6d,
	0x2c, 0x01, 0x02, 0x14, 0x11, 0x0a, 0x20, 0x1f,
	0x1e, 0x1d, 0x1c, 0x15, 0x30, 0x2e, 0x3d, 0x2b,
	0x0d, 0x6c, 0x3b, 0x17, 0x31, 0x32, 0x33, 0x2d,
	0x1a, 0x69, 0x6f, 0x12, 0x34, 0x35, 0x36, 0x2a,
	0x19, 0x70, 0x08, 0xfe, 0x37, 0x38, 0x39, 0x2f,
	0x29, 0xfe, 0xfe, 0xfe, 0xfe, 0x5e, 0x28, 0xfe
};

/*
	p^[\
*/
static inline void putpat(uint8 col, uint8 row, const uint8 *pat, uint8 len)
{
	memcpy(&vram[ROW_LCD2VRAM(row) * vramWidth + col * cellWidth], pat, len);
}

/*
	row
*/
static inline void clrline(uint8 row)
{
	memset(&vram[ROW_LCD2VRAM(row) * vramWidth], 0, vramWidth - 1);
}

/*
	ɃXN[
*/
void scrup(void)
{
	uint8 tmp;

	tmp = vram[vramWidth * 7 + lcdWidth];
	vram[vramWidth * 7 + lcdWidth] = vram[vramWidth * 6 + lcdWidth];
	vram[vramWidth * 6 + lcdWidth] = vram[vramWidth * 5 + lcdWidth];
	vram[vramWidth * 5 + lcdWidth] = vram[vramWidth * 4 + lcdWidth];
	vram[vramWidth * 4 + lcdWidth] = vram[vramWidth * 3 + lcdWidth];
	vram[vramWidth * 3 + lcdWidth] = vram[vramWidth * 2 + lcdWidth];
	vram[vramWidth * 2 + lcdWidth] = vram[vramWidth * 1 + lcdWidth];
	vram[vramWidth * 1 + lcdWidth] = vram[vramWidth * 0 + lcdWidth];
	vram[vramWidth * 0 + lcdWidth] = tmp;

	clrline(0);
	switch(machine) {
	case MACHINE_E200:
		memory[0x790d] = (memory[0x790d] + 1) % 8;
		z80outport(NULL, 0x58, (memory[0x790d] << 3) | 0xc0);
		break;
	case MACHINE_G815:
		memory[0x790d] = (memory[0x790d] + 1) % 8;
		z80outport(NULL, 0x50, (memory[0x790d] << 3) | 0xc0);
		break;
	case MACHINE_G850:
		clrline(6);
		clrline(7);
		memory[0x790d] = (memory[0x790d] + 1) % vramRows;
		z80outport(NULL, 0x40, (memory[0x790d] * 8) % (vramRows * 8) | 0x40);
		break;
	}
}

/*
	ɃXN[
*/
static void scrdown(uint8 row, uint8 col)
{
	int n;
	uint8 r;

	n = lcdWidth - row * cellWidth;
	for(r = vramRows - 1; r != row; r--) {
		memcpy(
		&vram[ROW_LCD2VRAM(r)     * vramWidth + col * cellWidth],
		&vram[ROW_LCD2VRAM(r - 1) * vramWidth + col * cellWidth],
		n);
	}
	clrline(r);
}

/*
	ʑŜ
*/
static void clrall(void)
{
	uint8 row;

	for(row = 0; row != vramCols; row++)
		clrline(row);
}

/*
	ŏ̕\
*/
void putchr(uint8 col, uint8 row, uint8 chr)
{
	putpat(col, row, font[chr], cellWidth);
}

/*
	̍s߂
*/
int nextRow(uint8 *col, uint8 *row)
{
	*col = 0;
	if(*row >= lcdRows - 1) {
		*row = lcdRows - 1;
		return TRUE;
	} else {
		(*row)++;
		return FALSE;
	}
}

/*
	̗߂
*/
int nextCol(uint8 *col, uint8 *row)
{
	if(*col >= lcdCols - 1)
		return nextRow(col, row);
	else {
		(*col)++;
		return FALSE;
	}
}

/*
	O̗߂
*/
void prevCol(uint8 *col, uint8 *row)
{
	if(*col == 0) {
		if(*row == 0)
			return;
		*col = lcdCols - 1;
		(*row)--;
	} else
		(*col)--;
}

/*
	2ڈȍ~̕\
*/
int putchrNext(uint8 *col, uint8 *row, uint8 chr)
{
	if(nextCol(col, row)) {
		scrup();
		putchr(*col, *row, chr);
		return TRUE;
	} else {
		putchr(*col, *row, chr);
		return FALSE;
	}
}

/*
	\
*/
static void putstr(uint8 col, uint8 row, char *str, ...)
{
	va_list v;
	char buf[256], *p;

	va_start(v, str);
	vsprintf(buf, str, v);
	va_end(v);

	for(p = buf; *p != 0; p++)
		putchr(col++, row, *p);
}

/*
	LCDɃhbg邩ׂ
*/
static uint8 point(int16 x, int16 y)
{
	/*
	printf("POINT(%d,%d)\n", x, y);
	fflush(stdout);
	*/

	if(x < 0 || y < 0 || x >= lcdWidth || y >= lcdHeight)
		return 0;

	return vram[ROW_LCD2VRAM(y / 8) * vramWidth + x];
}

/*
	LCDɓ_`
*/
static void pset(int16 x, int16 y, uint8 mode)
{
	uint8 mask, *p;

	if(x < 0 || y < 0 || x >= lcdWidth || y >= lcdHeight)
		return;

	p = &vram[ROW_LCD2VRAM(y / 8) * vramWidth + x];
	mask = 1 << (y % 8);

	switch(mode) {
	case 0:
		*p &= ~mask;
		break;
	case 1:
		*p |= mask;
		break;
	default:
		*p ^= mask;
		break;
	}
}

/*
	16bit̕ϐ
*/
static inline void swap16(void *x, void *y)
{
	uint16 tmp;

	tmp = *(uint16 *)x;
	*(uint16 *)x  = *(uint16 *)y;
	*(uint16 *)y  = tmp;
}

/*
	`̃NbsOs
*/
static void clip(int16 *x1, int16 *y1, int16 *x2, int16 *y2)
{
	*x1 = (*x1 < 0 ? 0: (*x1 >= lcdWidth  ? lcdWidth  - 1: *x1));
	*y1 = (*y1 < 0 ? 0: (*y1 >= lcdHeight ? lcdHeight - 1: *y1));
	*x2 = (*x2 < 0 ? 0: (*x2 >= lcdWidth  ? lcdWidth  - 1: *x2));
	*y2 = (*y2 < 0 ? 0: (*y2 >= lcdHeight ? lcdHeight - 1: *y2));
	if(*x1 > *x2) 
		swap16(x1, x2);
	if(*y1 > *y2)
		swap16(y1, y2);
}

/*
	LCDɐ`
*/
static void line(uint16 x1, uint16 y1, uint16 x2, uint16 y2, uint8 mode)
{
	int dx, dx0, dy, dy0, e;
	int16 x, y;

	/*
	printf("LINE(%d,%d)-(%d,%d),%d\n", x1, y1, x2, y2, mode & 0x03);
	fflush(stdout);
	*/

	dx0 = x2 - x1; dx = (dx0 > 0 ? dx0: -dx0);
	dy0 = y2 - y1; dy = (dy0 > 0 ? dy0: -dy0);

	if(dx > dy) {
		if(dx0 < 0) {
			swap16(&x1, &x2);
			swap16(&y1, &y2);
			dy0 = -dy0;
		}
		for(x = x1, y = y1, e = 0; x <= x2; x++) {
			e += dy;
			if(e > dx) {
				e -= dx;
				y += (dy0 > 0 ? 1: -1);
			}
			pset(x, y, mode);
		}
	} else {
		if(dy0 < 0) {
			swap16(&x1, &x2);
			swap16(&y1, &y2);
			dx0 = -dx0;
		}
		for(y = y1, x = x1, e = 0; y <= y2; y++) {
			e += dx;
			if(e > dy) {
				e -= dy;
				x += (dx0 > 0 ? 1: -1);
			}
			pset(x, y, mode);
		}
	}
}

/*
	LCDɎlp`
*/
static void box(int16 x1, int16 y1, int16 x2, int16 y2, uint8 mode)
{
	int16 i;

	/*
	printf("BOX(%d,%d)-(%d,%d),%d\n", x1, y1, x2, y2, mode);
	fflush(stdout);
	*/

	if(x1 > x2)
		swap16(&x1, &x2);
	if(y1 > y2)
		swap16(&y1, &y2);

	for(i = x1; i <= x2; i++)
		pset(i, y1, mode);
	for(i = y1 + 1; i <= y2 - 1; i++)
		pset(x1, i, mode);
	if(x1 != x2)
		for(i = y1 + 1; i <= y2 - 1; i++)
			pset(x2, i, mode);
	if(y1 != y2)
		for(i = x1; i <= x2; i++)
			pset(i, y2, mode);
}

/*
	LCDɓhԂlp`
*/
static void boxfill(int16 x1, int16 y1, int16 x2, int16 y2, uint8 mode)
{
	uint16 i, j;

	/*
	printf("BOXFILL(%d,%d)-(%d,%d),%d\n", x1, y1, x2, y2, mode);
	fflush(stdout);
	*/

	clip(&x1, &y1, &x2, &y2);
	for(j = y1; j <= y2; j++)
		for(i = x1; i <= x2; i++)
			pset(i, j, mode);
}

/*
	LCDɃp^[`
*/
static void gprint(int16 x, int16 y, uint8 pat)
{
	/*
	printf("GPRINT(%d,%d),%02x\n", x, y, pat);
	fflush(stdout);
	*/

	pset(x, y - 7, (pat & 0x01 ? 1: 0));
	pset(x, y - 6, (pat & 0x02 ? 1: 0));
	pset(x, y - 5, (pat & 0x04 ? 1: 0));
	pset(x, y - 4, (pat & 0x08 ? 1: 0));
	pset(x, y - 3, (pat & 0x10 ? 1: 0));
	pset(x, y - 2, (pat & 0x20 ? 1: 0));
	pset(x, y - 1, (pat & 0x40 ? 1: 0));
	pset(x, y - 0, (pat & 0x80 ? 1: 0));
}

/*
	ĂL[𓾂(waitȂ)
*/
uint8 getKeycode(void)
{
	const static uint8 zero[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	int len;
	uint8 k, *p;

	if(keyBreak)
		return GKEY_BREAK;
	if(memcmp(keyMatrix, zero, 10) == 0)
		return 0;

	for(p = keyMatrix, k = 1; *p == 0; p++, k += 8)
		;
	switch(*p) {
	case 0x01:
		break;
	case 0x02:
		k += 1; break;
	case 0x04:
		k += 2; break;
	case 0x08:
		k += 3; break;
	case 0x10:
		k += 4; break;
	case 0x20:
		k += 5; break;
	case 0x40:
		k += 6; break;
	case 0x80:
		k += 7; break;
	default:
		return GKEY_DOUBLE;
	}

	len = (int )(&keyMatrix[9] - p);
	if(len != 0 && memcmp(p + 1, zero, len) != 0)
		return GKEY_DOUBLE;
	return k | (keyShift ? 0x80: 0);
}

/*
	ĂL[𓾂(wait)
*/
uint8 getKeycodeWait(void)
{
	uint8 keycode;

	while(getKeycode() != 0) {
		updateKey();
		updateLCD();
		delay(1000 / freqUpdateIO);
	}
	while((keycode = getKeycode()) == 0) {
		updateKey();
		updateLCD();
		delay(1000 / freqUpdateIO);
	}
	return keycode;
}

/*
	L[R[hASCIIɕϊ
*/
static inline uint8 keycode2ascii(uint8 keycode, int is_large)
{
	if(is_large && keycode < 0x49)
		return keycode2ascii_l[keycode];
	else
		return keycode2ascii_s[keycode];
}

/*
	l(16i)𓾂
*/
static int ishex(uint8 ascii)
{
	return 
	ascii == 0x30 ||
	ascii == 0x31 ||
	ascii == 0x32 ||
	ascii == 0x33 ||
	ascii == 0x34 ||
	ascii == 0x35 ||
	ascii == 0x36 ||
	ascii == 0x37 ||
	ascii == 0x38 ||
	ascii == 0x39 ||
	ascii == 0x61 || ascii == 0x41 ||
	ascii == 0x62 || ascii == 0x42 ||
	ascii == 0x63 || ascii == 0x43 ||
	ascii == 0x64 || ascii == 0x44 ||
	ascii == 0x65 || ascii == 0x45 ||
	ascii == 0x66 || ascii == 0x46;
}
static void gethex(void *num, int length, uint8 col, uint8 row)
{
	int n, first = TRUE;
	char buf[8];
	uint8 k, x = length * 2;

	if(length == sizeof(uint8))
		putstr(col, row, "%02X", *(uint8 *)num);
	else
		putstr(col, row, "%04X", *(uint16 *)num);

	while((k = getKeycodeWait()) != GKEY_RETURN) {
		if(k == GKEY_CLS || first) {
			first = FALSE;
			while(x > 0)
				putchr(col + --x, row, 0x20);
		}
		if(ishex(keycode2ascii_l[k])) {
			if(x >= length * 2)
				x--;
			buf[x] = keycode2ascii_l[k];
			putchr(col + x, row, buf[x]);
			x++;
		}
	}
	buf[x] = 0;

	if(x == 0)
		return;
	sscanf(buf, "%x", &n);
	if(length == sizeof(uint8))
		*(uint8 *)num = n;
	else
		*(uint16 *)num = n;
}

/*
	WX^j
*/
static inline uint16 destroy16(void)
{
	static uint32 rnd = 0xffffffff;
	
	rnd = rnd * 65541 + 1;
	return rnd >> 16;
}
static inline uint8 destroy8(void)
{
	return destroy16() >> 8;
}

/*
	SWX^\
*/
static inline int iocs_0030(Z80stat *z)
{
	clrall();
	putstr(0, 0, "PC=%04X  AF=%02X %02X", z->r16.pc, z->r.a, z->r.f);
	putstr(0, 1, "SP=%04X  BC=%02X %02X", z->r16.sp, z->r.b, z->r.c);
	putstr(0, 2, "IX=%04X  DE=%02X %02X", z->r16.ix, z->r.d, z->r.e);
	putstr(0, 3, "IY=%04X  HL=%02X %02X", z->r16.iy, z->r.h, z->r.l);
	getKeycodeWait();

	return 1000;
}

/*
	荞ݏ
*/
static inline int iocs_0038(Z80stat *z)
{
	z->r.halt = 0;
	z->r.iff  = 3;
	interruptMask = 0x0f;

	return 1000;
}

/*
	҂ (PC-G850p)
*/
static inline int iocs_8aad(Z80stat *z)
{
	return 1500;
}

/*
	OtBbN (PC-G815p)
*/
static inline int iocs_02_f9f8(Z80stat *z)
{
	z->r.a = point(z->r16.hl, z->r16.de);
	z->r.c = 1 << (z->r16.de % 8);
	return 1000;
}
static inline int iocs_0d_c76e(Z80stat *z)
{
	pset(z->r16.hl, z->r16.de, z->m[0x7f0f]);
	return 1000;
}
static inline int iocs_0d_c5fc(Z80stat *z)
{
	int16 x = z->m[0x7968] << 8U | z->m[0x7967];
	int16 y = z->m[0x796a] << 8U | z->m[0x7969];

	line(z->r16.hl, z->r16.de, x, y, z->m[0x7f0f]);
	return 6000;
}
static inline int iocs_0d_c4a9(Z80stat *z)
{
	int16 x1 = z->r16.hl;
	int16 y1 = z->r16.de;
	int16 x2 = z->m[0x7968] << 8U | z->m[0x7967];
	int16 y2 = z->m[0x796a] << 8U | z->m[0x7969];

	z->r16.de = y2;
	box(x1, y1, x2, y2, z->m[0x7f0f]);
	return 8000;
}
static inline int iocs_0d_c442(Z80stat *z)
{
	int16 x1 = z->r16.hl;
	int16 y1 = z->r16.de;
	int16 x2 = z->m[0x7968] << 8U | z->m[0x7967];
	int16 y2 = z->m[0x796a] << 8U | z->m[0x7969];

	z->r16.de = y2;
	box(x1, y1, x2, y2, z->m[0x777f]);
	return 10000;
}
static inline int iocs_0d_c532(Z80stat *z)
{
	int16 x1 = z->r16.hl;
	int16 y1 = z->r16.de;
	int16 x2 = z->m[0x7968] << 8U | z->m[0x7967];
	int16 y2 = z->m[0x796a] << 8U | z->m[0x7969];

	z->r16.de = y2;
	boxfill(x1, y1, x2, y2, z->m[0x7f0f]);
	return 10000;
}
static inline int iocs_0d_c595(Z80stat *z)
{
	int16 x = z->m[0x7968] << 8U | z->m[0x7967];
	int16 y = z->m[0x796a] << 8U | z->m[0x7969];

	line(z->r16.hl, z->r16.de, x, y, z->m[0x777f]);
	return 6000;
}
static inline int iocs_02_f892(Z80stat *z)
{
	int16 x = z->m[0x79dc] << 8U | z->m[0x79db];
	int16 y = z->m[0x79de] << 8U | z->m[0x79dd];

	if(++(z->m[0x79db]) == 0)
		++(z->m[0x79dc]);
	z->r16.hl = x;
	z->r16.de = y & 0xff80;
	gprint(x, y, z->r.a);
	return 4000;
}
static inline int iocs_9490(Z80stat *z)
{
	uint16 address;
	uint8 page;

	if(useROM)
		return -1;

	page    = z->m[z->r16.pc + 3];
	address = z->m[z->r16.pc + 5] << 8U | z->m[z->r16.pc + 4];
	
	z->r16.pc = (z->m[z->r16.sp + 1] << 8U | z->m[z->r16.sp]) - 3;
	z->r16.sp += 2;

	switch(page) {
	case 0x02:
		switch(address) {
		case 0xf892:
			return iocs_02_f892(z);
		case 0xf9f8:
			return iocs_02_f9f8(z);
		}
		break;
	case 0x0d:
		switch(address) {
		case 0xc76e:
			return iocs_0d_c76e(z);
		case 0xc5fc:
			return iocs_0d_c5fc(z);
		case 0xc4a9:
			return iocs_0d_c4a9(z);
		case 0xc442:
			return iocs_0d_c442(z);
		case 0xc532:
			return iocs_0d_c532(z);
		case 0xc595:
			return iocs_0d_c595(z);
		}
		break;
	}

	/*
	printf("page=%02x address=%04x\n", page, address);
	fflush(stdout);
	*/

	return 1000;
}

/*
	OtBbN (PC-G850p)
*/
static inline int iocs_0e_ffca(Z80stat *z)
{
	z->r.a = point(z->r16.hl, z->r16.de);
	z->r.c = 1 << (z->r16.de % 8);
	return 1000;
}
static inline int iocs_0d_ffd0(Z80stat *z)
{
	pset(z->r16.hl, z->r16.de, z->m[0x777f]);
	return 1000;
}
static inline int iocs_0d_ffd3(Z80stat *z)
{
	int16 x = z->m[0x7968] << 8U | z->m[0x7967];
	int16 y = z->m[0x796a] << 8U | z->m[0x7969];

	line(z->r16.hl, z->r16.de, x, y, z->m[0x777f]);
	return 5000;
}
static inline int iocs_0d_ffd6(Z80stat *z)
{
	int16 x1 = z->r16.hl;
	int16 y1 = z->r16.de;
	int16 x2 = z->m[0x7968] << 8U | z->m[0x7967];
	int16 y2 = z->m[0x796a] << 8U | z->m[0x7969];
	
	z->r16.de = y2;
	box(x1, y1, x2, y2, z->m[0x777f]);
	return 6000;
}
static inline int iocs_0d_ffd9(Z80stat *z)
{
	int16 x1 = z->r16.hl;
	int16 y1 = z->r16.de;
	int16 x2 = z->m[0x7968] << 8U | z->m[0x7967];
	int16 y2 = z->m[0x796a] << 8U | z->m[0x7969];
	
	z->r16.de = y2;
	boxfill(x1, y1, x2, y2, z->m[0x777f]);
	return 10000;
}
static inline int iocs_0e_ffa3(Z80stat *z)
{
	int16 x = z->m[0x79dc] << 8U | z->m[0x79db];
	int16 y = z->m[0x79de] << 8U | z->m[0x79dd];

	if(++(z->m[0x79db]) == 0)
		++(z->m[0x79dc]);
	z->r16.hl = x;
	z->r16.de = y & 0xff80;
	gprint(x, y, z->r.a);
	return 3000;
}
static inline int iocs_bb6b(Z80stat *z)
{
	uint16 address;
	uint8 page;

	if(useROM)
		return -1;

	page    = z->m[z->r16.pc + 3];
	address = z->m[z->r16.pc + 5] << 8U | z->m[z->r16.pc + 4];

	z->r16.pc = (z->m[z->r16.sp + 1] << 8U | z->m[z->r16.sp]) - 3;
	z->r16.sp += 2;

	switch(page) {
	case 0x0d:
		switch(address) {
		case 0xc76e:
		case 0xffd0:
			return iocs_0d_ffd0(z);
		case 0xc595:
		case 0xffd3:
			return iocs_0d_ffd3(z);
		case 0xc442:
		case 0xffd6:
			return iocs_0d_ffd6(z);
		case 0xc4cb:
		case 0xffd9:
			return iocs_0d_ffd9(z);
		}
		break;
	case 0x0e:
		switch(address) {
		case 0xca08:
		case 0xffca:
			return iocs_0e_ffca(z);
		case 0xc92e:
		case 0xffa3:
			return iocs_0e_ffa3(z);
		}
		break;
	}

	/*
	printf("page=%02x address=%04x\n", page, address);
	fflush(stdout);
	*/

	return 1000;
}

/*
	ĂL[ASCIIR[h𓾂(wait) (PC-G850p)
*/
static inline int iocs_bcc4(Z80stat *z)
{
	z->r.a = keycode2ascii_l[getKeycodeWait()];
	z->r.f |= 0x01;

	z->r.f |= destroy8() & 0xfe;
	z->r.b = 0;
	z->r.c = destroy8();
	z->r16.hl   = destroy16();
	z->r16.bc_d = destroy16();
	z->r16.de_d = destroy16();
	z->r16.hl_d = destroy16();

	return 100000;
}

/*
	ĂL[𓾂(waitȂ)
*/
static inline int iocs_be53(Z80stat *z)
{
	z->r.a = getKeycode();
	if(z->r.a) 
		z->r.f |= 0x01;	/* ? */
	else
		z->r.f &= ~0x01;	/* ? */
	
	if(z->r.f & 0x01)
		z->r.b = destroy8();
	z->r.f |= destroy8() & 0xfe;
	z->r16.bc_d = destroy16();
	z->r16.de_d = destroy16();
	z->r16.hl_d = destroy16();

	switch(machine) {
	case MACHINE_E200:
	case MACHINE_G815:
		return 18000;
	case MACHINE_G850:
	default:
		return 30000;
	}
}

/*
	L[R[hASCIIR[hɕϊ
*/
static inline int iocs_be56(Z80stat *z)
{
	if(z->m[0x78f0] & 0x08) {
		z->r.b = z->r.a;
		z->r.f = 0x10;
	} else {
		z->r.a = keycode2ascii(z->r.a, z->m[0x7901] & 0x02);
		z->r.f = 0x44;
	}

	return 500;
}

/*
	1\(L܂)
*/
static inline int iocs_be5f(Z80stat *z)
{
	if(z->r.e >= vramCols || z->r.d >= vramRows)
		return 100;
	
	putchr(z->r.e, z->r.d, z->r.a);

	z->r16.af = destroy16();
	z->r.b   = 0;
	z->r.c   = destroy8();
	z->r16.hl = destroy16();

	return 1800;
}

/*
	1\(L܂܂Ȃ)
*/
static inline int iocs_be62(Z80stat *z)
{
	if(z->r.e >= vramCols || z->r.d >= vramRows)
		return 100;
	
	putchr(z->r.e, z->r.d, (z->r.a > 0x20 ? z->r.a: 0x20));

	z->r16.af = destroy16();
	z->r.b   = 0;
	z->r.c   = destroy8();
	z->r16.hl = destroy16();
	
	return 1800;
}

/*
	ɃXN[
*/
static inline int iocs_be65(Z80stat *z)
{
	if(z->r.e >= vramCols || z->r.d >= vramRows)
		return 100;

	scrdown(z->r.e, z->r.d);

	z->r16.af = destroy16();
	z->r16.bc = 0x0000;
	z->r16.de = destroy16();
	z->r16.hl = destroy16();
	
	return 5000;
}

/*
	ĂL[𓾂(wait)
*/
static inline int iocs_bcfd(Z80stat *z)
{
	z->r.a = getKeycodeWait();
	z->r.f |= 0x01;

	z->r.f |= destroy8() & 0xfe;
	z->r16.bc_d = destroy16();
	z->r16.de_d = destroy16();
	z->r16.hl_d = destroy16();

	return 100000;
}

/*
	16i2̃L[͂𓾂
*/
static inline int iocs_bd09(Z80stat *z)
{
	clrall();
	putstr(0, 0, "A=");
	gethex(&z->r.a, sizeof(z->r.a), 2, 0);
	clrall();

	z->r16.bc_d = destroy16();
	z->r16.de_d = destroy16();
	z->r16.hl_d = destroy16();

	return 100000;
}

/*
	16i4̃L[͂𓾂
*/
static inline int iocs_bd0f(Z80stat *z)
{
	clrall();
	putstr(0, 0, "HL=");
	gethex(&z->r16.hl, sizeof(z->r16.hl), 3, 0);
	clrall();

	z->r16.af   = 0x0000;
	z->r16.bc_d = destroy16();
	z->r16.de_d = destroy16();
	z->r16.hl_d = destroy16();

	return 100000;
}

/*
	p^[\
*/
static inline int iocs_bfd0(Z80stat *z)
{
	int state;
	uint8 n;

	if(z->r.e >= vramCols || z->r.d >= vramRows || z->r.b == 0)
		return 100;

	n = z->r.e + z->r.b / cellWidth; n = (n < vramCols ? n: vramCols);
	putpat(z->r.e, z->r.d, &z->m[z->r16.hl], z->r.b);

	state = 100 + 170 * z->r.b;
	z->r.a = z->m[z->r16.hl];
	z->r.e += z->r.b;
	z->r16.hl += z->r.b - 1;
	z->r.b = 0;

	z->r.f = destroy8();

	return state;
}

/*
	ɃXN[
*/
static inline int iocs_bfeb(Z80stat *z)
{
	scrup();

	z->r16.af = 0x0044;
	z->r.b = 0;
	z->r16.hl = destroy16();
	
	return 5000;
}

/*
	n̕\
*/
static inline int iocs_bfee(Z80stat *z)
{
	int state;

	if(z->r.e >= vramCols || z->r.d >= vramRows || z->r.b == 0)
		return 100;
	
	state = 100 + 1800 * z->r.b;
	
	putchr(z->r.e, z->r.d, z->r.a);
	while(--z->r.b)
		putchrNext(&z->r.e, &z->r.d, z->r.a);
	
	z->r.a   = 0;
	z->r.f   = destroy8();
	z->r16.hl = destroy16();
	
	return state;
}

/*
	\
*/
static inline int iocs_bff1(Z80stat *z)
{
	int state;

	if(z->r.e >= vramCols || z->r.d >= vramRows || z->r.b == 0)
		return 100;

	state = 100 + 1800 * z->r.b;

	z->r.c = 0;
	putchr(z->r.e, z->r.d, z->m[z->r16.hl]);
	while(--z->r.b)
		if(putchrNext(&z->r.e, &z->r.d, z->m[++z->r16.hl]))
			z->r.c++;

	z->r16.af = destroy16();

	return state;
}

/*
	d؂
*/
static inline int iocs_c110(Z80stat *z)
{
	ioReset = 0;
	z->r.iff  = 0;
	z->r.halt = 1;
	return 0;
}

/*
	IOCSR[G~[g
*/
int z80subroutine(Z80stat *z, uint16 address)
{
	if(address < 0x8000 && address > 0x0040)
		return -1;

	switch(address) {
	case 0x0030:
		return iocs_0030(z);
	case 0x0038:
		return iocs_0038(z);
	case 0xbcfd:
		return iocs_bcfd(z);
	case 0xbe53:
		return iocs_be53(z);
	case 0xbe56:
		return iocs_be56(z);
	case 0xbe5f:
		return iocs_be5f(z);
	case 0xbe62:
		return iocs_be62(z);
	case 0xbe65:
		return iocs_be65(z);
	case 0xbd09:
		return iocs_bd09(z);
	case 0xbd0f:
		return iocs_bd0f(z);
	case 0xbfd0:
		return iocs_bfd0(z);
	case 0xbfeb:
		return iocs_bfeb(z);
	case 0xbfee:
		return iocs_bfee(z);
	case 0xbff1:
		return iocs_bff1(z);
	case 0xc110:
		return iocs_c110(z);
	}
	
	switch(machine) {
	case MACHINE_E200:
		break;
	case MACHINE_G815:
		switch(address) {
		case 0x93cd:
		case 0x9490:
			return iocs_9490(z);
		}
		break;
	case MACHINE_G850:
		switch(address) {
		case 0x8aad:
			return iocs_8aad(z);
		case 0x93cb:
		case 0x93cd:
		case 0xbb6b:
			return iocs_bb6b(z);
		case 0xbcc4:
			return iocs_bcc4(z);
		}
		break;
	}

	if(address >= 0x8000) {
#ifdef WARN_UNKNOWN_IO
		printf("UNKNOWN CALL %04x\n", address);
#endif
		return (useROM ? -1: 100);
	} else
		return -1;
}

/*
	Copyright 2005 ~ 2008 maruhiro
	All rights reserved. 

	Redistribution and use in source and binary forms, 
	with or without modification, are permitted provided that 
	the following conditions are met: 

	 1. Redistributions of source code must retain the above copyright notice, 
	    this list of conditions and the following disclaimer. 

	 2. Redistributions in binary form must reproduce the above copyright notice, 
	    this list of conditions and the following disclaimer in the documentation 
	    and/or other materials provided with the distribution. 

	THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
	FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
	THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* eof */
