/*
	SHARP PC-G800 series emulator
	[j^ƃj[
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <limits.h>
#include <stdarg.h>
#include <dirent.h>
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32)
#	include <windows.h>
#endif
#include "g800.h"

#define REPEAT_START_DELAY	200	/* L[ŏɉĂ當o܂ł̑҂ */
#define REPEAT_DELAY	100	/* oԊu */

/* L[R[h->ϊ\ */
static const char key2ascii_normal[] = {
	  0,   0, 'q', 'w', 'e', 'r', 't', 'y',
	'u', 'a', 's', 'd', 'f', 'g', 'h', 'j',
	'k', 'z', 'x', 'c', 'v', 'b', 'n', 'm',
	',',   0,   0,   0,   0,   0, ' ',   0,
	  0,   0,   0,   0, '0', '.', '=', '+',
	  0, 'l', ';',   0, '1', '2', '3', '-',
	  0, 'i', 'o',   0, '4', '5', '6', '*',
	  0, 'p',   0,   0, '7', '8', '9', '/',
	')',   0,   0,   0,   0, '^', '(',   0,
	  0,   0,   0,   0,   0,   0,   0,   0,
	  0,   0,   0
};
static const char key2ascii_caps[] = {
	  0,   0, 'Q', 'W', 'E', 'R', 'T', 'Y',
	'U', 'A', 'S', 'D', 'F', 'G', 'H', 'J',
	'K', 'Z', 'X', 'C', 'V', 'B', 'N', 'M',
	',',   0,  0,    0,   0,   0, ' ',   0,
	  0,   0,  0,    0, '0', '.', '=', '+',
	  0, 'L', ';',   0, '1', '2', '3', '-',
	  0, 'I', 'O',   0, '4', '5', '6', '*',
	  0, 'P',   0,   0, '7', '8', '9', '/',
	')',   0,   0,   0,   0, '^', '(',   0,
	  0,   0,   0,   0,   0,   0,   0,   0,
	  0,   0,   0
};
static const char key2ascii_shift[] = {
	    0,   0, '!', '"', '#', '$', '%', '&',
	 '\'', '[', ']', '{', '}', '\\', '|', '~',
	  '_',   0,   0,   0,   0,    0,   0,   0,
	  '?',   0,   0,   0,   0,    0,   0,   0,
	    0,   0,   0,   0,   0,    0,   0,   0,
	    0, '=', ':',   0,   0,    0,   0,   0,
	    0, '<', '>',   0,   0,    0,   0,   0,
	    0, '@',   0,   0,   0,    0,   0,   0,
	    0,   0,   0,   0,   0,    0,   0,   0,
	    0,   0,   0,   0,   0,    0,   0,   0,
	    0,   0,   0
};

/* J[\ʒu */
static uint8 curCol;
static uint8 curRow;

/* VRAMۑ̈ */
static uint8 savedVram[166 * 9];
static uint8 savedLcdContrast;
static uint8 savedLcdTop;
static uint8 savedLcdBegin;
static uint8 savedMemory790d;
static int savedLcdScales;

/*
	16i̕𐔒lɕϊ ()
*/
static int atoix(const char *buf)
{
	int x;

	sscanf(buf, "%x", &x);
	return x;
}

/*
	WX^ݒ肷
*/
void go(uint16 pc)
{
	z80.r16.af = 0x0044;
	z80.r16.bc = 0x0000;
	z80.r16.de = 0x0000;
	z80.r16.hl = 0x0100;
	z80.r16.ix = 0x7c05;
	z80.r16.iy = 0x7c03;
	z80.r16.sp = 0x7ff6;
	z80.r16.pc = pc;
}

/*
	vOǂݍ
*/
int loadProg(uint16 *begin_ret, const char *path)
{
	int size;
	off_t begin;

	if(path == NULL || strcmp(path, "") == 0)
		return 0;
	if((size = readHex(path, memory, &begin, 0x8000, FALSE)) == 0)
		return -1;

	if(begin_ret != NULL)
		*begin_ret = begin;
	return size;
}

/*
	VRAMޔ ()
*/
static void pushVram(void)
{
	memcpy(savedVram, vram, sizeof(vram));
	savedLcdContrast = lcdContrast;
	savedLcdTop = lcdTop;
	savedLcdBegin = lcdBegin;
	savedMemory790d = memory[0x790d];
	savedLcdScales = lcdScales;

	lcdContrast = 0x0f;
	lcdScales = 2;
	updateLCDContrast();
}

/*
	VRAM𕜋A ()
*/
static void popVram(void)
{
	memcpy(vram, savedVram, sizeof(vram));
	lcdContrast = savedLcdContrast;
	lcdTop = savedLcdTop;
	lcdBegin = savedLcdBegin;
	memory[0x790d] = savedMemory790d;
	lcdScales = savedLcdScales;
	updateLCDContrast();
}

/*
	\ʒu߂ ()
*/
static void glocate(uint8 col, uint8 row)
{
	curCol = col;
	curRow = row;
}

/*
	\ ()
*/
static void gprintf(const char *str, ...)
{
	va_list v;
	char buf[256], *p;

	va_start(v, str);
	vsprintf(buf, str, v);
	va_end(v);

	for(p = buf; *p != 0; p++)
		switch(*p) {
		case '\n':
			if(nextRow(&curCol, &curRow))
				scrup();
			break;
		case '\r':
			curCol = 0;
			break;
		default:
			putchr(curCol, curRow, *p);
			if(nextCol(&curCol, &curRow) && *(p + 1) != 0)
				scrup();
			break;
		}

	/*updateLCD();*/
}

/*
	ʑŜ ()
*/
static void gcls(void)
{
	glocate(0, lcdRows - 1);
	gprintf("\n\n\n\n\n\n\n\n");
	lcdTop = lcdBegin = memory[0x790d] = 0;
	glocate(0, 0);
}

/*
	L[𗣂܂ő҂()
*/
static void waitRelease(void)
{
	do {
		updateKey();
		delay(1000 / freqUpdateIO);
	} while(getKeycode() != GKEY_NONE);
}

/*
	L[R[h𓾂(I[gs[g) ()
*/
static uint8 _getKeycodeRepeat(void)
{
	static int repeat_counter = 0;
	static int repeat_start_counter = 0;
	static uint8 old_key;
	static uint8 last_key = GKEY_NONE;

	updateKey();

	old_key = last_key;
	last_key = getKeycode();
	if(last_key != old_key || last_key == GKEY_NONE || last_key == GKEY_DOUBLE) {
		repeat_counter = REPEAT_DELAY * freqUpdateIO / 1000;
		repeat_start_counter = REPEAT_START_DELAY * freqUpdateIO / 1000;
	}

	if(last_key != old_key)
		return last_key;
	if(repeat_start_counter-- > 0)
		return GKEY_NONE;
	if(repeat_counter-- > 0)
		return GKEY_NONE;
	repeat_counter = REPEAT_DELAY * freqUpdateIO / 1000;
	return last_key;
}
static uint8 getKeycodeRepeat(void)
{
	uint8 key;

	while(key = _getKeycodeRepeat(), key == GKEY_NONE || key == GKEY_DOUBLE) {
		updateLCD();
		delay(1000 / freqUpdateIO);
	}
	
	updateLCD();
	return key;
}

static uint8 getKeycodeRepeatNowait(void)
{
	uint8 key;

	key = _getKeycodeRepeat();
	updateLCD();
	delay(1000 / freqUpdateIO);

	return key;
}

/*
	L[R[h𕶎ɕϊ(_ggets̉)
*/
static char key2ascii(uint8 key, int caps, int shift)
{
	if(shift)
		return key2ascii_shift[key];
	else if(caps)
		return key2ascii_caps[key];
	else
		return key2ascii_normal[key];
}

/*
	𓾂(obt@Ȃ) (ggets̉)
*/
static uint8 _ggets(char *buf)
{
	uint8 x, y, k;
	char *p;

	for(p = buf; *p != 0; p++) {
		putchr(curCol, curRow, *p);
		if(nextCol(&curCol, &curRow))
			scrup();
	}

	*p = 0;
	for(;;) {
		x = curCol;
		y = curRow;
		glocate(x, y);
		gprintf("_");
		k = getKeycodeRepeat() & 0x7f;
		glocate(x, y);
		gprintf(" ");
		glocate(x, y);

		switch(k) {
		case GKEY_CLS:
			for(; p > buf; p--) {
				prevCol(&curCol, &curRow);
				putchr(curCol, curRow, ' ');
			}
			strcpy(buf, "");
			break;
		case GKEY_TAB:
		case GKEY_OFF:
		case GKEY_BREAK:
			return k;
		case GKEY_CAPS:
			memory[0x7901] ^= 0x02;
			break;
		case GKEY_RETURN:
			if(nextRow(&curCol, &curRow))
				scrup();
			return k;
		case GKEY_BACKSPACE:
		case GKEY_LEFT:
			if(p == buf)
				return k;
			prevCol(&curCol, &curRow);
			putchr(curCol, curRow, ' ');
			*--p = 0;
			break;
		default:
			if(p == buf + lcdCols * lcdRows)
				break;
			*p = key2ascii(k, memory[0x7901] & 0x02, keyShift);
			if(*p == 0)
				break;
			putchr(curCol, curRow, *p);
			if(nextCol(&curCol, &curRow))
				scrup();
			*++p = 0;
			break;
		}
	}
}

/*
	𓾂 ()
*/
static uint8 ggets(char *buf)
{
	strcpy(buf, "");
	return _ggets(buf);
}

/*
	16i̕? ()
*/
static int ishex(char c)
{
	return
	isdigit(c) ||
	c == 'a' || c == 'A' ||
	c == 'b' || c == 'B' ||
	c == 'c' || c == 'C' ||
	c == 'd' || c == 'D' ||
	c == 'e' || c == 'E' ||
	c == 'f' || c == 'F';
}

/*
	l(16i2oCg)𓾂 ()
*/
static int gethex16(uint8 col, uint8 row, const char *prompt)
{
	char buf[256], *p;

	glocate(col, row);
	gprintf("%*s", lcdCols - col, " ");
	glocate(col, row);
	gprintf("%s", prompt);
	if(ggets(buf) != GKEY_RETURN)
		return -1;

	if(strlen(buf) == 0 || strlen(buf) > 4)
		return -1;
	for(p = buf; *p != 0; p++)
		if(!ishex(*p))
			return -1;
	return atoix(buf);
}

/*
	bZ[W\
*/
void popup(const char *icon, const char *str, ...)
{
	va_list v;
	int count = 0;
	char buf[256];

	va_start(v, str);
	vsprintf(buf, str, v);
	va_end(v);

	pushVram();
	waitRelease();
	
	glocate(1, 0);
	gprintf(buf);
	do {
		glocate(0, 0);
		count++;
		if(count < freqUpdateIO / 2)
			gprintf("%s", icon);
		else if(count < freqUpdateIO)
			gprintf(" ");
		else
			count = 0;
		updateLCD();
		updateKey();
		delay(1000 / freqUpdateIO);
	} while(getKeycode() == GKEY_NONE);

	waitRelease();
	popVram();
}

/*
	ꗗ獀ڂI ()
*/
static int selectItem(char **sel, char *buf)
{
	uint8 key;
	int top = 0, cur = 0, i, n, y;
	char *item[256], *p;

	for(n = 0, p = buf; n < sizeof(item) / sizeof(item[0]) && *p != 0; n++, p += strlen(p) + 1)
		item[n] = p;

	for(;;) {
		gcls();
		for(i = 0, y = 0; i < lcdRows; i++, y++) {
			glocate(1, y);
			gprintf("%.12s", (top + i < n ? item[top + i]: ""));
		}
		for(i = lcdRows, y = 0; i < lcdRows * 2; i++, y++) {
			glocate(14, y);
			gprintf("%.10s", (top + i < n ? item[top + i]: ""));
		}
		glocate(0, cur);
		gprintf("\x13");

		switch(key = getKeycodeRepeat()) {
		case GKEY_UP:
			if(cur != 0)
				cur--;
			else if(top != 0) {
				cur = lcdRows - 1;
				top -= lcdRows;
			}
			break;
		case GKEY_LEFT:
			if(top != 0)
				top -= lcdRows;
			break;
		case GKEY_DOWN:
			if(top + cur + 1 >= n)
				;
			else if(cur != lcdRows - 1)
				cur++;
			else if(top + lcdRows < n) {
				cur = 0;
				top += lcdRows;
			}
			break;
		case GKEY_RIGHT:
			if(top + lcdRows + cur < n)
				top += lcdRows;
			break;
		case GKEY_CLS:
		case GKEY_TAB:
		case GKEY_BREAK:
		case GKEY_OFF:
			*sel = NULL;
			return key;
		case GKEY_RETURN:
		case GKEY_SPACE:
			*sel = item[top + cur];
			return key;
		}
	}
}

/*
	ꗗɃt@Cǉ ()
*/
static int addItem(char *list, int size, const char *item)
{
	int len;
	char *p, *q, p_last, item_last;

	for(p = list; *p != 0; p += strlen(p) + 1) {
		p_last = p[strlen(p) - 1];
		item_last = item[strlen(item) - 1];
		if(
		(p_last == '/' && item_last == '/') ||
		(p_last != '/' && item_last != '/')
		) {
			if(strcasecmp(p, item) > 0)
				break;
		} else {
			if(item_last == '/')
				break;
		}
	}
	for(q = p; *q != 0; q += strlen(q) + 1)
		;
	
	len = strlen(item);
	if(list + size < q + len + 2)
		return FALSE;

	memmove(p + len + 1, p, (int )(q - p));
	strcpy(p, item);
	return TRUE;
}

/*
	fBNg𓾂()
*/
static char *getDir(char *path)
{
	char *p;

	for(p = path + strlen(path) - 1; p >= path && *p != '/'; p--)
		*p = 0;

	return path;
}

/*
	sԍ܂IntelHEX`ł邩`FbN ()
*/
static int checkFile(const char *path)
{
	FILE *fp;
	int i;
	char buf[256];

	if((fp = fopen(path, "r")) == NULL)
		return FALSE;
	
	for(i = 0; i < 4 && !feof(fp); i++) {
		fgets(buf, sizeof(buf), fp);
		if(!isdigit(buf[0]) && buf[0] != ':' && buf[0] != 0x1a) {
			fclose(fp);
			return FALSE;
		}
	}

	fclose(fp);
	return TRUE;
}

/*
	t@C̈ꗗ𓾂 (selectFilẻ)
*/
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32)
static int makeFileList(char *file, int size, const char *dir_name)
{
	HANDLE h;
	WIN32_FIND_DATA found;
	int n = 0;
	char base_name[PATH_MAX], buf[PATH_MAX], *p;

	memset(file, 0, size);

	sprintf(base_name, "%s*.*", dir_name);
	for(p = base_name; *p != 0; p++)
		if(*p == '/')
			*p = '\\';

	if((h = FindFirstFile(base_name, &found)) == INVALID_HANDLE_VALUE)
		return -1;
	do {
		if(strcmp(found.cFileName, ".") == 0)
			continue;
		if(found.dwFileAttributes & (FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_TEMPORARY))
			continue;
		if(found.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			sprintf(buf, "%s/", found.cFileName);
		else {
			if(useFileFilter) {
				getDir(strcpy(buf, dir_name));
				strcat(buf, found.cFileName);
				if(!checkFile(buf))
					continue;
			}
			sprintf(buf, "%s", found.cFileName);
		}

		if(!addItem(file, size, buf))
			break;
		n++;
	} while(FindNextFile(h, &found));

	FindClose(h);

	return n;
}
#else
static int makeFileList(char *file, int size, const char *dir_name)
{
	DIR *dir;
	struct dirent *p;
	int n = 0;
	char buf[PATH_MAX];

	memset(file, 0, size);

	if((dir = opendir(dir_name)) == NULL)
		return -1;
	for(p = readdir(dir); p != NULL; p = readdir(dir)) {
		if((p->d_name[0] == '.' && strcmp(p->d_name, "..") != 0))
			continue;
		if(p->d_type == DT_DIR)
			sprintf(buf, "%s/", p->d_name);
		else {
			if(useFileFilter) {
				getDir(strcpy(buf, dir_name));
				strcat(buf, p->d_name);
				if(!checkFile(buf))
					continue;
			}
			sprintf(buf, "%s", p->d_name);
		}

		if(!addItem(file, size, buf))
			break;
		n++;
	}
	closedir(dir);

	return n;
}
#endif

/*
	̃fBNgɈړ (selectFilẻ)
*/
static int upDir(char *path)
{
	char *p;

	for(p = path + strlen(path) - 1; p >= path && *p != '/'; p--)
		*p = 0;

	p -= 2;
	if(p < path || strcmp(p, "../") == 0) {
		strcat(path, "../");
		return FALSE;
	}

	for(; p >= path && *p != '/'; p--)
		*p = 0;
	return TRUE;
}

/*
	t@C̈ꗗI ()
*/
static uint8 _selectFile(char *path)
{
	char file[0x4000], *p = 0;
	uint8 key = GKEY_NONE;

	for(;;) {
		if(makeFileList(file, sizeof(file), (strcmp(path, "") != 0 ? path: "./")) > 0)
			key = selectItem(&p, file);
		else
			return GKEY_TAB;

		switch(key) {
		case GKEY_CLS:
			strcpy(path, "");
			return key;
		case GKEY_TAB:
		case GKEY_OFF:
			return key;
		default:
			if(p == NULL)
				return key;
			if(strcmp(p, "../") == 0)
				upDir(path);
			else if(p[strlen(p) - 1] == '/')
				strcat(path, p);
			else {
				getDir(path);
				strcat(path, p);
				return key;
			}
		}
	}
}
static int selectFile(char *path)
{
	uint8 key;
	char buf[PATH_MAX];

	strcpy(buf, path);

	for(;;) {
		gcls();
		gprintf("FILE:");
		key = _ggets(buf);
		if(key != GKEY_TAB)
			break;
		key = _selectFile(buf);
		if(key != GKEY_TAB)
			break;
	}
	gcls();

	if(key == GKEY_RETURN || key == GKEY_SPACE) {
		strcpy(path, buf);
		return TRUE;
	} else
		return FALSE;
}

/*
	\ł镶ɕϊ (dump0̉)
*/
static uint8 to_disp(uint8 chr)
{
	if(chr < 0x10 || chr > 0xf8)
		return '.';
	else
		return chr;
}

/*
	dump1ʕ\ (dump̉)
*/
static void dump0(uint16 address)
{
	uint8 sum = 0, *p;

	address &= 0xfff0;
	for(p = &memory[address]; p < &memory[address + 16]; p++)
		sum += *p;

	glocate(0, 0);
	gprintf(
	"%04X : %02X %02X %02X %02X  %c%c%c%c"
	"(%02X)   %02X %02X %02X %02X  %c%c%c%c"
	"       %02X %02X %02X %02X  %c%c%c%c"
	"       %02X %02X %02X %02X  %c%c%c%c",
	address, memory[address + 0], memory[address + 1], memory[address + 2], memory[address + 3], to_disp(memory[address + 0]), to_disp(memory[address + 1]), to_disp(memory[address + 2]), to_disp(memory[address + 3]),
	sum, memory[address + 4], memory[address + 5], memory[address + 6], memory[address + 7], to_disp(memory[address + 4]), to_disp(memory[address + 5]), to_disp(memory[address + 6]), to_disp(memory[address + 7]),
	memory[address + 8], memory[address + 9], memory[address + 10], memory[address + 11], to_disp(memory[address + 8]), to_disp(memory[address + 9]), to_disp(memory[address + 10]), to_disp(memory[address + 11]),
	memory[address + 12], memory[address + 13], memory[address + 14], memory[address + 15], to_disp(memory[address + 12]), to_disp(memory[address + 13]), to_disp(memory[address + 14]), to_disp(memory[address + 15])
	);

	if(lcdRows < 6)
		return;
	if(address >= 0xfff0) {
		gprintf(
		"------------------------"
		"                        "
		);
		return;
	}

	sum = 0;
	for(p = &memory[address + 16]; p < &memory[address + 32]; p++)
		sum += *p;

	gprintf(
	"%04X : %02X %02X %02X %02X  %c%c%c%c"
	"(%02X)   %02X %02X %02X %02X  %c%c%c%c",
	address + 16, memory[address + 16], memory[address + 17], memory[address + 18], memory[address + 19], to_disp(memory[address + 16]), to_disp(memory[address + 17]), to_disp(memory[address + 18]), to_disp(memory[address + 19]),
	sum, memory[address + 20], memory[address + 21], memory[address + 22], memory[address + 23], to_disp(memory[address + 20]), to_disp(memory[address + 21]), to_disp(memory[address + 22]), to_disp(memory[address + 23])
	);
}

/*
	Dump\ ()
*/
static uint8 dump(uint16 address)
{
	uint8 k;

	for(;;) {
		dump0(address);
		switch(k = getKeycodeRepeat()) {
		case GKEY_UP:
			address -= 16;
			break;
		case GKEY_DOWN:
			address += 16;
			break;
		case GKEY_CLS:
		case GKEY_OFF:
		case GKEY_BREAK:
			gcls();
			return k;
		}
	}
}

/*
	ҏW ()
*/
static int edit0(uint16 address, int cur, uint8 k)
{
	uint8 x, *p;

	if(k == GKEY_0)
		x = 0;
	else if(k == GKEY_1)
		x = 1;
	else if(k == GKEY_2)
		x = 2;
	else if(k == GKEY_3)
		x = 3;
	else if(k == GKEY_4)
		x = 4;
	else if(k == GKEY_5)
		x = 5;
	else if(k == GKEY_6)
		x = 6;
	else if(k == GKEY_7)
		x = 7;
	else if(k == GKEY_8)
		x = 8;
	else if(k == GKEY_9)
		x = 9;
	else if(k == GKEY_A)
		x = 0xa;
	else if(k == GKEY_B)
		x = 0xb;
	else if(k == GKEY_C)
		x = 0xc;
	else if(k == GKEY_D)
		x = 0xd;
	else if(k == GKEY_E)
		x = 0xe;
	else if(k == GKEY_F)
		x = 0xf;
	else
		return FALSE;

	p = &memory[address + cur / 2];
	if(cur % 2 == 0)
		*p = (*p & 0x0f) | (x << 4);
	else
		*p = (*p & 0xf0) | x;
	return TRUE;
}
static uint8 edit(uint16 address)
{
	const uint8 col[] = {
		7, 8, 10, 11, 13, 14, 16, 17,
		7, 8, 10, 11, 13, 14, 16, 17,
		7, 8, 10, 11, 13, 14, 16, 17,
		7, 8, 10, 11, 13, 14, 16, 17
	};
	const uint8 row[] = {
		0, 0, 0, 0, 0, 0, 0, 0,
		1, 1, 1, 1, 1, 1, 1, 1,
		2, 2, 2, 2, 2, 2, 2, 2,
		3, 3, 3, 3, 3, 3, 3, 3
	};
	int cur = (address & 0x0f) * 2;
	int count = 0;
	uint8 k;

	for(;;) {
		/* XN[ */
		if(cur < 0) {
			cur += 32;
			address -= 16;
		} else if(cur >= 32) {
			cur -= 32;
			address += 16;
		}
		address &= 0x7ff0;

		/* L[𓾂, J[\\ */
		count = 0;
		do {
			if(count == 0) {
				dump0(address);
				glocate(col[cur], row[cur]);
				gprintf("\x10");
			} else if(count == freqUpdateIO / 2) {
				dump0(address);
			}
			count = (count + 1) % freqUpdateIO;
		} while(k = getKeycodeRepeatNowait(), k == GKEY_NONE);

		/* s */
		switch(k) {
		case GKEY_UP:
			cur -= 8;
			break;
		case GKEY_DOWN:
			cur += 8;
			break;
		case GKEY_LEFT:
			cur--;
			break;
		case GKEY_RIGHT:
			cur++;
			break;
		default:
			if(edit0(address, cur, k))
				cur++;
			break;
		case GKEY_CLS:
		case GKEY_OFF:
		case GKEY_BREAK:
			gcls();
			return k;
		}
	}
}

static uint8 disasm(uint16 address)
{
	char buf[256];
	uint8 k, row, *p;

	for(;;) {
		gcls();
		p = &memory[address];
		for(row = 0; row < lcdRows; row++) {
			p = z80disasm(buf, p);
			glocate(0, row);
			gprintf("%04X:%s", (int )(p - memory), buf);
		}

		switch(k = getKeycodeRepeat()) {
		case GKEY_DOWN:
			address = (int )((uint8 *)z80disasm(buf, &memory[address]) - memory);
			break;
		case GKEY_UP:
			break;
		case GKEY_CLS:
			gcls();
			return k;
		}
	}
}

/*
	t@Cj[\ (menủ)
*/
static int menuFile(void)
{
	int size, start, end;
	uint16 address;
	char path[PATH_MAX];
	char info[256] = "";

	for(;;) {
		/* j[\ */
		gcls();
		gprintf(
		"      <<<< FILE >>>\n"
		"\n"
		"  Read  Write\n"
		"\n"
		"\n"
		);
		glocate(0, lcdRows - 2);
		gprintf("%s", info);
		
		/* s */
		switch(getKeycodeRepeat()) {
		case GKEY_R:
			/* ǂݍ */
			strcpy(path, "");
			if(selectFile(path) && strcmp(path, "") != 0) {
				if((size = loadProg(&address, path)) <= 0)
					sprintf(info, "READ ERROR");
				else
					sprintf(info, "READ OK:%04X-%04X", address, address + size - 1);
			}
			break;
		case GKEY_W:
			/*  */
			if((start = gethex16(0, lcdRows - 2, "START:")) < 0)
				break;
			if((end = gethex16(10, lcdRows - 2, "END:")) < 0)
				break;
			strcpy(path, "");
			if(!selectFile(path))
				break;
			if(writeHex(path, memory, start, end - start + 1) <= 0) {
				sprintf(info, "WRITE ERROR");
				break;
			}
			sprintf(info, "WRITE OK:%04X-%04X", start, end);
			break;
		case GKEY_CLS:
		case GKEY_BREAK:
			/* ߂ */
			waitRelease();
		case GKEY_BASIC:
		case GKEY_TEXT:
		case GKEY_OFF:
			return TRUE;
		}
	}
}

/*
	SIOj[s (menủ)
*/
static int menuSio(void)
{
	char path[PATH_MAX];

	for(;;) {
		/* j[SIOԂ\ */
		gcls();
		gprintf(
		"       <<< SIO >>>\n"
		"\n"
		"  In   Out   Switch\n"
		"\n"
		"\n"
		);
		glocate(0, lcdRows - 2);
		switch(sioMode) {
		case SIO_MODE_STOP:
			gprintf("MODE:STOP");
			break;
		case SIO_MODE_IN:
			gprintf("MODE:IN[%s]", pathSioIn);
			break;
		case SIO_MODE_OUT:
			gprintf("MODE:OUT[%s]", pathSioOut);
			break;
		}

		/* s */
		switch(getKeycodeRepeat()) {
		case GKEY_S:
			/* SIO؂ւ */
			sioMode = (sioMode < 2 ? sioMode + 1: 0);
			if(sioMode == SIO_MODE_IN && strcmp(pathSioIn, "") == 0)
				sioMode = SIO_MODE_OUT;
			if(sioMode == SIO_MODE_OUT && strcmp(pathSioOut, "") == 0)
				sioMode = SIO_MODE_STOP;
			break;
		case GKEY_I:
			/* SIO */
			strcpy(path, "");
			if(selectFile(path))
				sioLoad(path);
			break;
		case GKEY_O:
			/* SIOo */
			strcpy(path, "");
			if(selectFile(path))
				sioSave(path);
			break;
		case GKEY_CLS:
		case GKEY_BREAK:
			/* ߂ */
			waitRelease();
		case GKEY_BASIC:
		case GKEY_TEXT:
		case GKEY_OFF:
			return TRUE;
		}
	}
}

/*
	j[s
*/
int menu(void)
{
	pushVram();
	gcls();

	for(;;) {
		/* j[\ */
		gcls();
		gprintf(
		"      *** MENU ***\n"
		"\n"
		"  Sio  File  CLS:Return"
		);

		/* s */
		switch(getKeycodeRepeat()) {
		case GKEY_F:
			/* t@Cj[ */
			if(menuFile())
				goto last;
			break;
		case GKEY_S:
			/* SIOj[ */
			if(menuSio())
				goto last;
			break;
		case GKEY_CLS:
		case GKEY_BREAK:
			/* ߂ */
			waitRelease();
		case GKEY_BASIC:
		case GKEY_TEXT:
		case GKEY_OFF:
			goto last;
		}
	}

last:;
	popVram();
	return FALSE;
}

/*
	p[^𓾂 (getParam1, getParam2̉)
*/
static const char *_getParam(const char *p, int *param)
{
	char tmp[256];
	const char *q;
	char *r;

	for(q = p, r = tmp; ishex(*q); q++, r++)
		*r = *q;
	*r = 0;
	
	if(p == q) {
		*param = -1;
		return (*q == 0 || *q == ',' ? q: NULL);
	}
	
	*param = atoix(tmp);
	return q;
}

/*
	1̃p[^𓾂 (monitor̉)
*/
static int getParam1(const char *buf, int off, int *param)
{
	return (_getParam(buf + off, param) != NULL);
}

/*
	2̃p[^𓾂 (monitor̉)
*/
static int getParam2(const char *buf, int off, int *param1, int *param2)
{
	const char *p;

	if((p = _getParam(buf + off, param1)) == NULL)
		return FALSE;
	if(*p != ',')
		return FALSE;
	if(_getParam(p + 1, param2) == NULL)
		return FALSE;
	return TRUE;
}

/*
	[j^s
*/
void monitor(void)
{
	int size, param1, param2;
	uint16 address;
	uint8 key;
	char buf[256], path[PATH_MAX];

	/* ^Cg\ */
	gcls();
	gprintf("MACHINE LANGUAGE MONITOR");

	for(;;) {
		/* vvg\ */
		gprintf("*");
		strcpy(path, "");

		/* ͂ꂽR}h𓾂 */
		key = ggets(buf);
		switch(key) {
		case GKEY_RETURN:
			break;
		case GKEY_CLS:
			gcls();
			continue;
		case GKEY_OFF:
			return;
		default:
			gprintf("\n");
			continue;
		}

		/* R}hs */
		if(buf[0] == 0) {
			/* R}hȂ */
		} else if(memcmp(buf, "D", 1) == 0 && getParam1(buf, 1, &param1)) {
			/* D:_v */
			dump(param1 > 0 ? param1: 0x100);
		} else if(memcmp(buf, "E", 1) == 0 && getParam1(buf, 1, &param1)) {
			/* E:ҏW */
			edit(param1 > 0 ? param1: 0x100);
		} else if(memcmp(buf, "U", 1) == 0 && getParam1(buf, 1, &param1)) {
			/* U:tAZu */
			disasm(param1 > 0 ? param1: z80.r16.pc);
		} else if(memcmp(buf, "G", 1) == 0 && getParam1(buf, 1, &param1) && param1 >= 0) {
			/* G:s */
			go(param1);
			exec(TRUE);
		} else if(memcmp(buf, "R", 1) == 0) {
			/* R:ǂݍ */
			if((size = loadProg(&address, pathSioIn)) <= 0)
				gprintf("READ ERROR\n");
			else
				gprintf("INFO:%04X-%04X\n", address, address + size - 1);
		} else if(memcmp(buf, "W", 1) == 0 && getParam2(buf, 1, &param1, &param2) && param1 >= 0 && param2 >= 0) {
			/* W: */
			if(writeHex(pathSioOut, memory, param1, param2 - param1) <= 0)
				gprintf("WRITE ERROR\n");
		} else if(memcmp(buf, "USER", 4) == 0 && getParam1(buf, 4, &param1)) {
			/* USER:[UGA */
			if(param1 >= 0x100 && param1 <= 0x7fff) {
				memory[0x7fff] = (param1 + 1) >> 8;
				memory[0x7ffe] = (param1 + 1) & 0xff;
			}
			size = (memory[0x7fff] * 0x100 + memory[0x7ffe] - 1) & 0xffff;
			if(size == 0xffff)
				gprintf("FREE:NOT RESERVED\n");
			else
				gprintf("FREE:%04X-%04X\n", 0x100, size);
		} else {
			/* G[ */
			gprintf(" SYNTAX ERROR\n");
		}
	}
}

/*
	Copyright 2005 ~ 2008 maruhiro
	All rights reserved. 

	Redistribution and use in source and binary forms, 
	with or without modification, are permitted provided that 
	the following conditions are met: 

	 1. Redistributions of source code must retain the above copyright notice, 
	    this list of conditions and the following disclaimer. 

	 2. Redistributions in binary form must reproduce the above copyright notice, 
	    this list of conditions and the following disclaimer in the documentation 
	    and/or other materials provided with the distribution. 

	THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
	FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
	THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* eof */
